import numpy as np
import sys
# # temp solution for directory.
sys.path.append("./src/")
import math
from undefined.UDFunction import UDFunction
from undefined.GraphGenerator import UDGraph
from undefined.Utils import UDPrimitive, check_division_by_zero, check_log, check_pow, check_arc



def cos(udobject):
    """calculate the cosine operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function

    Raises:

        TypeError:raised if input is not compatiable with cosine operation

    Returns:
        if input is udfunction object,update val and der by cosine operation.
        if input is UDGraph object,update notes and function by cosine operation.
        if input is int,float,ndarray object,update them in cosine operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        if isinstance(udobject._val, (int, float)):
            new_val = math.cos(udobject._val)
            new_der = - 1 * math.sin(udobject._val) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.cos(udobject._val)
            new_der = -1 * np.sin(udobject._val) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    elif isinstance(udobject, UDGraph):
        new_func = UDPrimitive.COS
        if isinstance(udobject._val, (int, float)):
            new_val = math.cos(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.cos(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        return np.cos(udobject)

    elif isinstance(udobject, (int, float)):
        return math.cos(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def sin(udobject):
    """calculate the sin operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:

        TypeError:raised if input is not compatiable with sin operation

    Returns:
        if input is udfunction object,update val and der by sin operation.
        if input is UDGraph object,update notes and function by sin operation.
        if input is int,float,ndarray object,update them in sin operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        if isinstance(udobject._val, (int, float)):
            new_val = math.sin(udobject._val)
            new_der = math.cos(udobject._val) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.sin(udobject._val)
            new_der = np.cos(udobject._val) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)
    elif isinstance(udobject, UDGraph):
        new_func = UDPrimitive.SIN
        if isinstance(udobject._val, (int, float)):
            new_val = math.sin(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.sin(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        return np.sin(udobject)

    elif isinstance(udobject, (int, float)):
        return math.sin(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def tan(udobject):
    """calculate the tangent operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:

        TypeError:raised if input is not compatiable with tangent operation

    Returns:
        if input is udfunction object,update val and der by tangent operation.
        if input is UDGraph object,update notes and function by tangent operation.
        if input is int,float,ndarray object,update them in tangent operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        if isinstance(udobject._val, (int, float)):
            check_division_by_zero(math.cos(udobject._val))
            new_val = math.tan(udobject._val)
            new_der = (1 / (math.cos(udobject._val)) ** 2) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            check_division_by_zero(np.cos(udobject._val))
            new_val = np.tan(udobject._val)
            new_der = (1 / (np.cos(udobject._val)) ** 2) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)
    elif isinstance(udobject, UDGraph):
        new_func = UDPrimitive.TAN
        if isinstance(udobject._val, (int, float)):
            check_division_by_zero(math.cos(udobject._val))
            new_val = math.tan(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            check_division_by_zero(np.cos(udobject._val))
            new_val = np.tan(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        check_division_by_zero(np.cos(udobject))
        return np.tan(udobject)

    elif isinstance(udobject, (int, float)):
        check_division_by_zero(math.cos(udobject))
        return math.tan(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def sinh(udobject):
    """calculate the sinh operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        The result from the sinh operation. 
    """
    return (exp(udobject) - exp(-udobject)) / 2


def cosh(udobject):
    """calculate the cosh operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        The result from the cosh operation. 
    """
    return (exp(udobject) + exp(-udobject)) / 2


def tanh(udobject):
    """calculate the tanh operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        The result from the tanh operation. 
    """
    return sinh(udobject) / cosh(udobject)


def coth(udobject):
    """calculate the coth operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        The result from the coth operation. 
    """
    return cosh(udobject) / sinh(udobject)


def sech(udobject):
    """calculate the sech operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        The result from the sech operation. 
    """
    return 1 / cosh(udobject)


def csch(udobject):
    """calculate the csch operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        The result from the csch operation. 
    """
    return 1 / sinh(udobject)


def arccos(udobject):
    """calculate the arccos operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:
        TypeError:raised if input is not compatiable with arccos operation

    Returns:
        if input is udfunction object,update val and der by arccos operation.
        if input is UDGraph object,update notes and function by arccos operation.
        if input is int,float,ndarray object,update them in arccos operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        check_arc(udobject._val)
        if isinstance(udobject._val, (int, float)):
            new_val = math.acos(udobject._val)
            new_der = (-1 / math.sqrt(1 - udobject._val**2)) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.arccos(udobject._val)
            new_der = (-1 / np.sqrt(1 - udobject._val**2)) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    elif isinstance(udobject, UDGraph):
        check_arc(udobject._val)
        new_func = UDPrimitive.ACOS
        if isinstance(udobject._val, (int, float)):
            new_val = math.acos(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.arccos(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        check_arc(udobject)
        return np.arccos(udobject)

    elif isinstance(udobject, (int, float)):
        check_arc(udobject)
        return math.acos(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def arcsin(udobject):
    """calculate the arcsin operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:
        TypeError:raised if input is not compatiable with arcsin operation

    Returns:
        if input is udfunction object,update val and der by arcsin operation.
        if input is UDGraph object,update notes and function by arcsin operation.
        if input is int,float,ndarray object,update them in arcsin operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        check_arc(udobject._val)
        if isinstance(udobject._val, (int, float)):
            new_val = math.asin(udobject._val)
            new_der = (1 / math.sqrt(1 - udobject._val**2)) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.arcsin(udobject._val)
            new_der = (1 / np.sqrt(1 - udobject._val**2)) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)
    elif isinstance(udobject, UDGraph):
        check_arc(udobject._val)
        new_func = UDPrimitive.ASIN
        if isinstance(udobject._val, (int, float)):
            new_val = math.asin(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.arcsin(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        check_arc(udobject)
        return np.arcsin(udobject)

    elif isinstance(udobject, (int, float)):
        check_arc(udobject)
        return math.asin(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def arctan(udobject):
    """calculate the arctan operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:
        TypeError:raised if input is not compatiable with arctan operation

    Returns:
        if input is udfunction object,update val and der by arctan operation.
        if input is UDGraph object,update notes and function by arctan operation.
        if input is int,float,ndarray object,update them in arctan operation by their own types.
    """
    if isinstance(udobject, UDFunction):
        if isinstance(udobject._val, (int, float)):
            new_val = math.atan(udobject._val)
            new_der = (1 / (1 + udobject._val ** 2)) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.arctan(udobject._val)
            new_der = (1 / (1 + udobject._val ** 2)) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)
    elif isinstance(udobject, UDGraph):
        new_func = UDPrimitive.ATAN
        if isinstance(udobject._val, (int, float)):
            new_val = math.atan(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.arctan(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        return np.arctan(udobject)

    elif isinstance(udobject, (int, float)):
        return math.atan(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def sqrt(udobject):
    """calculate the square root operation of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:

        TypeError:raised if input is not compatiable with square root operation

    Returns:
        if input is udfunction object,update val and der by square root operation.
        if input is UDGraph object,update notes and function by square root operation.
        if input is int,float,ndarray object,update them in square root operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        check_pow(udobject._val, 0.5)
        if isinstance(udobject._val, (int, float)):
            new_val = math.sqrt(udobject._val)
            new_der = 0.5 * math.pow(udobject._val, -0.5) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.sqrt(udobject._val)
            new_der = 0.5 * np.power(udobject._val, -0.5) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    elif isinstance(udobject, UDGraph):
        check_pow(udobject._val, 0.5)
        new_func = UDPrimitive.SQRT
        if isinstance(udobject._val, (int, float)):
            new_val = math.sqrt(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.sqrt(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        check_pow(udobject, 0.5)
        return np.sqrt(udobject)

    elif isinstance(udobject, (int, float)):
        check_pow(udobject, 0.5)
        return math.sqrt(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def exp(udobject):
    """calculate the square exponential of input

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:

        TypeError:raised if input is not compatiable with exponential operation

    Returns:
        if input is udfunction object,update val and der by exponential operation.
        if input is UDGraph object,update notes and function by exponential operation.
        if input is int,float,ndarray object,update them in exponential operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        if isinstance(udobject._val, (int, float)):
            new_val = math.exp(udobject._val)
            new_der = math.exp(udobject._val) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.exp(udobject._val)
            new_der = np.exp(udobject._val) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    elif isinstance(udobject, UDGraph):
        new_func = UDPrimitive.EXP
        if isinstance(udobject._val, (int, float)):
            new_val = math.exp(udobject._val)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.exp(udobject._val)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        return udgraph

    elif isinstance(udobject, np.ndarray):
        return np.exp(udobject)

    elif isinstance(udobject, (int, float)):
        return math.exp(udobject)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")


def standard_logistic(udobject):
    """this is the function we calculate the standard logistic. 
    It is different than the log() function

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Returns:
        return the standard logistic results.
    """
    return 1 / (1 + exp(-udobject))


def log(udobject, base=math.e):
    """calculate the log of input.
    We can handle the any bases in this log. Users can pass in the base argument. 

    Args:
        udobject (udfunction object,UDGraph object,ndarray,ndarray,int,float): User defined function/number

    Raises:
        TypeError:raised if input is not compatiable with log operation

    Returns:
        if input is udfunction object,update val and der by log operation.
        if input is UDGraph object,update notes and function by log operation.
        if input is int,float,ndarray object,update them in log operation by their own types. 
    """
    if isinstance(udobject, UDFunction):
        check_log(udobject._val, base)
        if isinstance(udobject._val, (int, float)):
            new_val = math.log(udobject._val, base)
            new_der = 1 / (math.log(base) * udobject._val) * udobject._der
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.log(udobject._val)
            new_val = new_val / math.log(base)
            new_der = 1 / (math.log(base) * udobject._val) * udobject._der
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        return UDFunction(new_val, new_der)

    elif isinstance(udobject, UDGraph):
        check_log(udobject._val, base)
        new_func = UDPrimitive.LOG
        if isinstance(udobject._val, (int, float)):
            new_val = math.log(udobject._val, base)
        elif isinstance(udobject._val, np.ndarray):
            new_val = np.log(udobject._val) / math.log(base)
        else:
            raise TypeError("error raised by undefined: unsupported attribute type.")
        udgraph = UDGraph(new_val, new_func)
        udgraph._parents.append(udobject)
        udgraph._params["base"] = base
        return udgraph

    elif isinstance(udobject, np.ndarray):
        check_log(udobject, base)
        return np.log(udobject) / math.log(base)

    elif isinstance(udobject, (int, float)):
        check_log(udobject, base)
        return math.log(udobject, base)

    else:
        raise TypeError("error raised by undefined: unsupported attribute type.")




import sys

# # temp solution for directory.
sys.path.append("./src/")
from undefined.GraphGenerator import GraphGenerator
from undefined.UDFunction import UDFunction
from undefined.Calculator import *
import numpy as np
from types import LambdaType



def trace(f, mode='forward', seeds=None, plot=False, **kwargs):
    """This is the main function that the users will interact with.

    Args:
        f ([LambdaType]): [This is the user defined function. Use list if there are multiple input functions]
        mode (str, optional): Defaults to 'forward'. Determine with mode to use to calculate the derivatives.
        seeds ([type], optional): Defaults to None. Used to calculate the Jacobian matrix
        plot (bool, optional):  Defaults to False. Need to use in combination with "reverse" mode. If True, it will save the computational graph to your current working directory.

    Raises:
        Errors will raise if unlawful inputs are given.

    Returns:
        The function values and the derivatives results will be saved in a tuple as the output.
    """
    if type(f) is list:
        vals, ders = [], []
        for f_ in f:
            val, der = trace(f_, mode, seeds, plot, **kwargs)
            vals.append(val)
            ders.append(der)

        return np.array(vals), np.stack(ders)
    elif type(f) is not LambdaType:
        raise TypeError("error raised by undefined: cannot handle non lambda functions.")
    fix_len = -1
    num_variables = len(kwargs)
    variables = {}
    varnames = f.__code__.co_varnames
    if len(varnames) != num_variables:
        raise AttributeError("error raised by undefined: inconsistent number of variables.")
    if mode == 'forward':
        for i, varname in enumerate(varnames):
            variable = kwargs[varname]
            
            if isinstance(variable, np.ndarray):
                # 【【vector input】】
                if len(variable.shape) <= 1 or variable.shape[0] != 1:
                    raise TypeError(
                        f"error raised by undefined: only support vector inputs of shape (1,n), invalid input {varname} has shape {variable.shape}")
                # 【single vector input】
                if num_variables == 1:
                    if seeds is not None:
                        if type(seeds) not in (int, float):
                            raise TypeError("error raised by undefined: incorrect type of seed vector, expect int or float")
                        seed = seeds
                    else:
                        seed = 1
                    variables[varname] = UDFunction(
                        variable, np.ones(variable.shape) * seed)
                # 【multiple vector inputs】
                else:
                    # check if all the vector inputs has the same len
                    if fix_len != -1:
                        if fix_len != variable.shape[1]:
                            raise AttributeError("error raised by undefined: cannot handle multiple vector inputs with different lengths")
                    else:
                        fix_len = variable.shape[1]
                    if seeds is not None:
                        if type(seeds) is not np.ndarray:
                            raise TypeError("error raised by undefined: at least one seed vector incorrectly defined")
                        seed = seeds[i]
                        if type(seed) is not np.ndarray:
                            raise TypeError("error raised by undefined: at least one seed vector incorrectly defined")
                        if len(seed) != num_variables:
                            raise AttributeError(f"error raised by undefined: incorrect shape for seed vectors, expect ({num_variables, num_variables})")
                        seed_vector = seed.reshape(num_variables,1) * np.ones(
                        (num_variables, fix_len), dtype=int)
                    else:
                        seed_vector = np.zeros(
                        (num_variables, fix_len), dtype=int)
                        seed_vector[i, :] += 1
                    variables[varname] = UDFunction(variable, seed_vector)

            elif isinstance(variable, (int, float)):
                # 【【single scalar input】】
                if num_variables == 1:
                    if seeds is not None:
                        if type(seeds) not in (int, float):
                            raise TypeError("error raised by undefined: incorrect type of seed vector, expect int or float")
                        seed = seeds
                        variables[varname] = UDFunction(variable, seed)
                    else:
                        variables[varname] = UDFunction(variable)
                # 【【multiple scalar inputs】】
                else:
                    if seeds is not None:
                        if type(seeds) is not np.ndarray:
                            raise TypeError("error raised by undefined: incorrect type of seed vectors, expect numpy.ndarray")
                        seed = seeds[i]
                        if len(seed) != num_variables:
                            raise AttributeError(f"error raised by undefined: incorrect shape for seed vector, expect ({num_variables, num_variables})")
                        seed_vector = seed.reshape(num_variables,1)
                    else:
                        seed_vector = np.zeros((num_variables,1), dtype=int)
                        seed_vector[i,:] = 1
                    variables[varname] = UDFunction(variable, seed_vector)
            else:
                raise TypeError(
                    "error raised by undefined: variable type not in (int, float, np.ndarray).")

        f = f(**variables)
        return f.val, f.der.tolist() if type(f.der) is np.ndarray else f.der

    elif mode == 'reverse':
        for i, varname in enumerate(varnames):
            variable = kwargs[varname]
            if isinstance(variable, np.ndarray):
                if variable.shape[0] != 1 or variable.shape[0] != 1:
                    raise TypeError(
                        f"error raised by undefined: only support vector inputs of shape (1, ), invalid input {varname} has shape {variable.shape}")
                variables[varname] = UDGraph(variable, varname = varname)
            elif isinstance(variable, (int, float)):
                variables[varname] = UDGraph(variable, varname = varname)
            else:
                raise TypeError(
                    "error raised by undefined: variable type not in (int, float, np.ndarray).")
        g = f(**variables)
        # check format of seeds
        seeds_dic = {}
        if seeds is not None:
            if num_variables == 1:
                # single input
                if type(seeds) not in (int, float):
                    raise TypeError("error raised by undefined: incorrect type of seed vector, expect int or float")
                seeds_dic[varnames[i]] = {varnames[i]:seeds}
            else:
                # multiple inputs
                if type(seeds) is not np.ndarray:
                    raise TypeError("error raised by undefined: incorrect type of seed vectors, expect numpy.ndarray")
                if seeds.shape[0] != num_variables or seeds.shape[1] != num_variables :
                    raise AttributeError(f"error raised by undefined: incorrect shape for seed vector, expect ({num_variables, num_variables})")

                for i in range(num_variables):
                    seeds_dic[varnames[i]] = {}
                    for j in range(num_variables):
                        seeds_dic[varnames[i]][varnames[j]] = seeds[i][j]
        udgenerator = GraphGenerator(g, variables, seeds_dic)
        if plot:
            print(udgenerator.generate_str())
            udgenerator.generate_graph()

        res_der = []
        for var_name in variables.keys():
            res_var = udgenerator.generate_derivative(var_name)
            if type(res_var) is np.ndarray:
                res_der.append(res_var.tolist()[0])
            elif type(res_var) is not list:
                res_der.append([res_var])
            else:
                res_der.append(res_der)

        # 【single scalar input】
        if len(res_der) == 1 and len(res_der[0]) == 1:
            return g.val, res_der[0][0]
        return g.val, res_der
    else:
        raise AttributeError("error raised by undefined: unsupported mode.")





import sys
sys.path.append("./src/")
import math
from undefined.Calculator import sin, cos, exp
from undefined.UDFunction import UDFunction
import numpy as np
import unittest


class TestUDFunction(unittest.TestCase):

    # the setUp and tearDown are the methods to keep the code dry.
    def setUp(self):

        alpha = 2.0
        beta = 3.0

        a = 2.0
        x = UDFunction(a)

        # addition
        self.f1 = -2*x + beta
        self.f2 = beta + 2*x
        self.f3 = 5*x + 3*x
        self.f4 = x + 5*x + 3*x
        self.f5 = 5 + 2

        # substraction
        self.f6 = 2*x - beta
        self.f7 = beta - 2*x
        self.f8 = 5*x - 3*x
        self.f9 = x - 5*x - 3*x
        self.f10 = 5 - 2

        # multiplication
        self.f11 = 2*x + beta
        self.f12 = x*6 + beta
        self.f13 = beta + x*2
        self.f14 = beta + 4*x
        self.f15 = (x + 1) * (x - 2)

        # true divide
        self.f16 = 3*x/2
        self.f17 = (x+1)/(x-5)
        self.f20 = 4/3*x
        self.f24 = 343 / x

        # floor divide
        self.f18 = 3*x//4
        self.f19 = (x+1)//(x-5)
        self.f21 = 4//3*x
        self.f25 = 343 // x

        # power
        self.f22 = 4*x**3
        self.f23 = 4**2
        self.f50 = exp(sin(x)) - cos(x**0.5) * sin((cos(x)**2.0 + x**2.0)**0.5)
        self.f60 = 2**x

        # define extreme cases
        self.a = 2
        self.b = -1.11
        self.c = 99999.99
        self.d = -99999.99
        self.e = np.array([[self.a, self.b]])
        self.f = np.array([[self.a, self.b, self.c, self.d]])
        self.s = "str"
        self.af = UDFunction(self.a)
        self.bf = UDFunction(self.b)
        self.cf = UDFunction(self.c)
        self.df = UDFunction(self.d)
        self.ef = UDFunction(self.e)
        self.ff = UDFunction(self.f)
        self.g = math.pi / 2
        self.h = math.e

    # tearDown is used to delete output files if applicable.
    def tearDown(self):
        pass

    # the convention is to name the testing function as "test_<meaningful name>",
    # otherwise the unittest will not run.
    def test_add_integration(self):

        a = 2.0
        x = UDFunction(a)

        self.assertEqual(self.f1.val, -1.0)
        self.assertEqual(self.f1.der, -2.0)

        self.assertEqual(self.f2.val, 7.0)
        self.assertEqual(self.f2.der, 2.0)

        self.assertEqual(self.f3.val, 16.0)
        self.assertEqual(self.f3.der, 8.0)

        self.assertEqual(self.f4.val, 18.0)
        self.assertEqual(self.f4.der, 9.0)

        with self.assertRaises(Exception):
            self.f5.val
        with self.assertRaises(Exception):
            self.f5.der

        with self.assertRaises(Exception):
            self.f30 = "1" + 2*x

        with self.assertRaises(Exception):
            self.f40 = 2*x + "1"

    def test_add_only(self):
        self.assertEqual(round(self.a+self.b, 2), 0.89)
        result1 = self.af+self.bf
        self.assertEqual(result1.val, 0.89)
        self.assertEqual(result1.der, 2)
        self.assertEqual(self.af.val + 1, 3)
        self.assertEqual(self.af.der + 1, 2)
        result2 = self.ff + 1
        self.assertEqual(result2.val.tolist(), [
                         [3.0000000e+00, -1.1000000e-01, 1.0000099e+05, -9.9998990e+04]])
        self.assertEqual(result2.der, 1)

    def test_radd_only(self):
        self.assertEqual(1+self.af.val, 3)
        self.assertEqual(1+self.af.der, 2)
        result1 = self.af.__radd__(self.bf)
        self.assertEqual(result1.val, 0.89)
        result2 = 1 + self.ff
        self.assertEqual(result2.val.tolist(), [
                         [3.0000000e+00, -1.1000000e-01, 1.0000099e+05, -9.9998990e+04]])
        self.assertEqual(result2.der, 1)

    def test_radd_integration(self):
        a = 2.0
        x = UDFunction(a)
        y = UDFunction(a + 2)
        z = "1"
        a = y.__radd__(x)
        self.assertEqual(a.val, 6)
        self.assertEqual(a.der, 2)
        with self.assertRaises(Exception):
            x.__radd__(z)

    def test_sub_integration(self):

        a = 2.0
        x = UDFunction(a)

        self.assertEqual(self.f6.val, 1.0)
        self.assertEqual(self.f6.der, 2.0)

        self.assertEqual(self.f7.val, -1.0)
        self.assertEqual(self.f7.der, -2.0)

        self.assertEqual(self.f8.val, 4.0)
        self.assertEqual(self.f8.der, 2.0)

        self.assertEqual(self.f9.val, -14.0)
        self.assertEqual(self.f9.der, -7.0)

        with self.assertRaises(Exception):
            self.f10.val
        with self.assertRaises(Exception):
            self.f10.der

        with self.assertRaises(Exception):
            self.f31 = "1" - 2*x

        with self.assertRaises(Exception):
            self.f41 = 2*x - "1"

    def test_rsub_integration(self):
        a = 2.0
        x = UDFunction(a)
        y = UDFunction(a + 2)
        z = "1"
        a = y.__rsub__(x)
        self.assertEqual(a.val, -2)
        self.assertEqual(a.der, 0)
        with self.assertRaises(Exception):
            x.__rsub__(z)

    def test_sub_only(self):
        self.assertEqual(round(self.a-self.b, 2), 3.11)
        result1 = self.af-self.bf
        self.assertEqual(result1.val, 3.11)
        self.assertEqual(result1.der, 0)
        self.assertEqual(self.af.val - 1, 1)
        self.assertEqual(self.af.der - 1, 0)
        result2 = self.ff - 1
        self.assertEqual(result2.val.tolist(), [
                         [1.0, -2.11, 99998.99, -100000.99]])
        self.assertEqual(result2.der, 1)

    def test_rsub_only(self):
        self.assertEqual(1-self.af.val, -1)
        self.assertEqual(1-self.af.der, 0)
        result1 = self.af.__rsub__(self.bf)
        self.assertEqual(result1.val, -3.11)
        result2 = 1 - self.ff
        self.assertEqual(result2.val.tolist(), [
                         [-1.0, 2.11, -99998.99, 100000.99]])
        self.assertEqual(result2.der, -1)

    def test_mul_integration(self):

        a = 2.0
        x = UDFunction(a)

        self.assertEqual(self.f11.val, 7.0)
        self.assertEqual(self.f11.der, 2.0)

        self.assertEqual(self.f12.val, 15.0)
        self.assertEqual(self.f12.der, 6.0)

        self.assertEqual(self.f13.val, 7.0)
        self.assertEqual(self.f13.der, 2.0)

        self.assertEqual(self.f14.val, 11.0)
        self.assertEqual(self.f14.der, 4.0)

        with self.assertRaises(Exception):
            self.f32 = "2" * 4*x

        with self.assertRaises(Exception):
            self.f42 = 4*x * "2"

    def test_rmul_integration(self):
        a = 2.0
        x = UDFunction(a)
        y = UDFunction(a + 2)
        z = "1"
        a = y.__rmul__(x)
        self.assertEqual(a.val, 8)
        self.assertEqual(a.der, 6)
        with self.assertRaises(Exception):
            x.__rmul__(z)

    def test_mul_only(self):
        result1 = self.af*self.bf
        self.assertEqual(result1.val, -2.22)
        self.assertEqual(result1.der, 0.89)
        self.assertEqual(self.af.val * 2, 4)
        self.assertEqual(self.af.der * 2, 2)
        result2 = self.ff * 2
        self.assertEqual(result2.val.tolist(), [
                         [4.0, -2.22, 199999.98, -199999.98]])
        self.assertEqual(result2.der, 2)

    def test_rmul_only(self):
        result1 = self.af.__rmul__(self.bf)
        self.assertEqual(result1.val, -2.22)
        self.assertEqual(result1.der, 0.89)
        self.assertEqual(2 * self.af.val, 4)
        self.assertEqual(2 * self.af.der, 2)
        result2 = 2 * self.ff
        self.assertEqual(result2.val.tolist(), [
                         [4.0, -2.22, 199999.98, -199999.98]])
        self.assertEqual(result2.der, 2)

    def test_neg(self):
        a = 2.0
        x = UDFunction(a)

        self.assertEqual(str(x.__neg__()), "value: -2.0 \nderivative: -1")

    def test_truedev_integration(self):

        a = 2.0
        x = UDFunction(a)

        self.assertEqual(self.f16.val, 3)
        self.assertEqual(self.f16.der, 1.5)

        self.assertEqual(self.f17.val, -1.0)
        self.assertEqual(round(self.f17.der, 2), -0.67)

        self.assertEqual(round(self.f20.val, 2), 2.67)
        self.assertEqual(round(self.f20.der, 2), 1.33)

        self.assertEqual(round(self.f24.val, 2), 171.5)
        self.assertEqual(round(self.f24.der, 2), -85.75)

        with self.assertRaises(Exception):
            self.f35 = "2" / 4*x

        with self.assertRaises(Exception):
            self.f45 = 4*x / "2"

    def test_rtruediv_integration(self):
        a = 2.0
        x = UDFunction(a)
        y = UDFunction(a + 2)
        z = "1"
        a = y.__rtruediv__(x)
        self.assertEqual(a.val, 0.5)
        self.assertEqual(a.der, 0.125)
        with self.assertRaises(Exception):
            x.__rtruediv__(z)

    def test_truediv_only(self):
        result1 = self.af/self.bf
        self.assertEqual(result1.val, -1.8)
        self.assertEqual(result1.der, -2.524)
        self.assertEqual(self.af.val / 2, 1)
        self.assertEqual(self.af.der / 2, 0.5)
        result2 = self.ff / 2
        self.assertEqual(result2.val.tolist(), [
                         [1.0, -0.56, 50000.0, -50000.0]])
        self.assertEqual(result2.der, 0.5)

    def test_rtruediv_only(self):
        result1 = self.af.__rtruediv__(self.bf)
        self.assertEqual(result1.val, -0.56)
        self.assertEqual(result1.der, 0.778)
        self.assertEqual(2 / self.af.val, 1)
        self.assertEqual(2 / self.af.der, 2)
        result2 = 2 / self.ff
        self.assertEqual(result2.val.tolist(), [[1.0, -1.8, 0.0, -0.0]])
        self.assertEqual(result2.der.tolist(), [[-0.5, -1.623, -0, -0]])

    def test_floordev_integration(self):

        a = 2.0
        x = UDFunction(a)

        self.assertEqual(self.f18.val, 1.0)
        self.assertEqual(self.f18.der, 0)

        self.assertEqual(self.f19.val, -1.0)
        self.assertEqual(round(self.f19.der, 2), -1.0)

        self.assertEqual(round(self.f21.val, 2), 2)
        self.assertEqual(round(self.f21.der, 2), 1)

        self.assertEqual(round(self.f25.val, 2), 171)
        self.assertEqual(round(self.f25.der, 2), -86)

        with self.assertRaises(Exception):
            self.f36 = "2" // 4*x

        with self.assertRaises(Exception):
            self.f46 = 4*x // "2"

    def test_rfloordev_integration(self):
        a = 2.0
        x = UDFunction(a)
        y = UDFunction(a + 2)
        z = "1"
        a = y.__rfloordiv__(x)
        self.assertEqual(a.val, 0)
        self.assertEqual(a.der, 0)
        with self.assertRaises(Exception):
            x.__rfloordiv__(z)

    def test_floordev_only(self):
        result1 = self.af // self.bf
        self.assertEqual(result1.val, -2)
        self.assertEqual(result1.der, -3)
        self.assertEqual(self.af.val // 2, 1)
        self.assertEqual(self.af.der // 2, 0)
        result2 = self.ff // 2
        self.assertEqual(result2.val.tolist(), [
                         [1.0, -1.0, 49999.0, -50000.0]])
        self.assertEqual(result2.der, 0)

    def test_rfloordev_only(self):
        result1 = self.af.__rfloordiv__(self.bf)
        self.assertEqual(result1.val, -1)
        self.assertEqual(result1.der, 0)
        self.assertEqual(2 // self.af.val, 1)
        self.assertEqual(2 // self.af.der, 2)
        result2 = 2 // self.ff
        self.assertEqual(result2.val.tolist(), [[1.0, -2.0, 0.0, -1.0]])
        self.assertEqual(result2.der.tolist(), [[-1.0, -2.0, -1.0, -1.0]])

    def test_power(self):

        self.assertEqual(self.f22.val, 32)
        self.assertEqual(self.f22.der, 48)

        self.assertEqual(round(self.f50.val, 2), 2.34)
        self.assertEqual(round(self.f50.der, 2), -0.64)  # Value check

        with self.assertRaises(Exception):
            self.f23.val
        with self.assertRaises(Exception):
            self.f23.der

        # test other as ndarray
        a = 2.0
        x = UDFunction(a)
        y = UDFunction(np.array([a]))
        self.f51 = x**y
        self.assertEqual(self.f51.val, 4)
        self.assertEqual(self.f51.der, 6.773)

        x1 = UDFunction(np.array([[2]]))
        y1 = UDFunction(np.array([[2]]))

        self.f52 = x1 ** y1
        self.assertEqual(self.f52.val[0], 4.0)
        self.assertEqual(self.f52.der[0], 6.773)

        # raise value error
        x2 = UDFunction(np.array([1]))
        y2 = UDFunction(np.array([1, 2]))

        with self.assertRaises(ValueError) as context:
            x2 ** y2

        x3 = UDFunction(np.array(2))
        y3 = UDFunction(2)

        self.f53 = x3 ** y3
        self.assertEqual(self.f53.val, 4)
        self.assertEqual(self.f53.der, 6.773)

        self.assertEqual(self.f60.val, 4)
        self.assertEqual(self.f60.der, 2.773)

    def test_rpow(self):
        a = 2
        # test __rpow__
        # This pa
        x = UDFunction(a)

        self.f60 = 2 ** x
        self.assertEqual(self.f60.val, 4)
        self.assertEqual(self.f60.der, 2.773)

    def test_eq(self):
        self.assertTrue(self.f22 == 32)
        self.assertTrue(self.f19 == -1)
        self.assertTrue(round(self.f25, 2) == 171)
        self.assertTrue(round(self.f20, 2) == 2.67)

        self.assertTrue(self.f22 == self.f22)
        self.assertTrue(self.f11 == self.f13)

        with self.assertRaises(TypeError):
            self.f11 == "self.f13"

    def test_ne(self):
        self.assertTrue(self.f19 != -12)
        self.assertTrue(round(self.f25, 2) != 161)
        self.assertTrue(self.f22 != self.f19)

        with self.assertRaises(TypeError):
            self.f11 != "self.f13"

    def test_lt(self):
        self.assertTrue(self.f19 < 12)
        self.assertTrue(round(self.f25, 2) < 200)
        self.assertTrue(self.f19 < self.f25)

        with self.assertRaises(TypeError):
            self.f25 < "self.f19"

    def test_gt(self):
        self.assertTrue(self.f19 > -12)
        self.assertTrue(round(self.f25, 2) > -200)
        self.assertTrue(self.f25 > self.f19)

        with self.assertRaises(TypeError):
            self.f25 > "self.f19"

    def test_le(self):
        self.assertTrue(self.f19 <= 12)
        self.assertTrue(round(self.f25, 2) <= 200)
        self.assertTrue(self.f19 <= self.f25)

        self.assertTrue(self.f22 <= 32)
        self.assertTrue(self.f19 <= -1)

        with self.assertRaises(TypeError):
            self.f25 <= "self.f19"

    def test_ge(self):
        self.assertTrue(self.f19 >= -12)
        self.assertTrue(round(self.f25, 2) >= -200)
        self.assertTrue(self.f25 >= self.f19)

        self.assertTrue(self.f22 >= 32)
        self.assertTrue(self.f19 >= -1)

        with self.assertRaises(TypeError):
            self.f25 >= "self.f19"


# this will help to run the unittest directly.
if __name__ == "__main__":
    unittest.main()
